/* Main program for inflating a gzipped file. */

/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#include "config.h"
#include "inflate.h"

static void
PRINTF_STYLE(1, 2)  
werror(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
}

static void
PRINTF_STYLE(1, 2)
NORETURN
die(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  exit(EXIT_FAILURE);
}

#define READ_UINT32(p) \
  ((p)[0] | ((p) [1] << 8) | ((p) [2] << 16) | ((p) [3] << 24))
#define READ_UINT16(p) \
  ((p)[0] | ((p) [1] << 8))

#if 0
static int
input_byte(struct input *self)
{
  assert(self->bits == 0);
  return getc(self->f);
}

static int
input_bytes(struct input *self, unsigned length, unsigned char *buf)
{
  size_t res;
  assert(self->bits == 0);
  assert(length > 0);
  res = fread(buf, length, 1, self->f);

  return res == 1;
}
#endif

static int
dump_zstring(FILE *f)
{
  int c;

  while ( (c = getc(f)) > 0)
    {
      if (c == '\\')
	werror("\\\\");
      else if (isprint(c))
	werror("%c", c);
      else
	werror("\\x%02x", c);
    }
  printf("\n");

  return c == 0;
}

static void
process_gzip_header (FILE *f)
{
  unsigned char header[10];
  unsigned flags;
  time_t mtime;

  if (fread(header, sizeof(header), 1, f) != 1)
    die("EOF in gzip header.\n");
  if (header[0] != 31 && header[1] != 139)
    die("Not a gzip file.");
  if (header[2] != 8)
    die("Uknown compression method %x\n", header[2]);

  flags = header[3];
  mtime = READ_UINT32(header + 4);

  /* Time is 32 bit signed. Support 64 bit time_t. */
  if (mtime > 0 && (mtime & 0x80000000))
    mtime -= 0x100000000;

  if (mtime != 0)
    werror("Modification time: %s", ctime(&mtime));

  werror("Extra flags = %d, OS = %d\n", header[8], header[9]);

  werror("Type: %s\n", (flags & 1) ? "ASCII text" : "Binary");
  if (flags & 4)
    {
      /* Extra data. Ignore it. */
      unsigned char buf[100];
      int xlen;
      
      werror("Extra field present. Ignored.\n");
      if (fread(buf, 2, 1, f) != 1)
	die("EOF in gzip header.\n");
      
      xlen = READ_UINT16(buf);

      if (xlen < 0)
	die("Unexpected EOF.\n");

      while (xlen >= sizeof(buf))
	{
	  if (fread(buf, sizeof(buf), 1, f) != 1)
	    die("Unexpected EOF.\n");
	  xlen -= sizeof(buf);
	}
      if (xlen > 0)
	if (fread(buf, xlen, 1, f) != 1)
	  die("Unexpected EOF.\n");
    }
  if (flags & 8)
    {
      /* File name */
      werror("File name: ");
      if (!dump_zstring(f))
	die("Unexpected EOF.\n");
    }
  if (flags & 16)
    {
      /* Comment */
      werror("Comment: ");
      if (!dump_zstring(f))
	die("Unexpected EOF.\n");
    }
  if (flags & 2)
    {
      /* Header CRC 16. */
      unsigned char buf[2];

      werror("Header CRC present, and ignored (not checked.)\n");

      if (fread(buf, sizeof(buf), 1, f) != 1)
	die("Unexpected EOF.\n");      
    }
}

static void
process_gzip_trailer (const char *trailer UNUSED)
{
  werror("Ignoring CRC32 checksum (not checked).\n");
}

#define BUF_SIZE 1000

int
main(int argc UNUSED, char **argv UNUSED)
{
  struct inflate_ctx ctx;
  char input_buffer[BUF_SIZE];
  size_t size;
  size_t done;

  const char *trailer;
  process_gzip_header (stdin);

  inflate_init(&ctx);

  size = done = 0;
  do
    {
      unsigned output_length;
      const char *output;

      if (done == size && (ctx.status & INFLATE_FLAG_INPUT_EXPECTED))
	{
	  size = fread(input_buffer, 1, sizeof(input_buffer), stdin);
	  if (size < sizeof(input_buffer) && ferror(stdin))
	    die("Read error.\n");
	  
	  done = 0;
	}

      done += inflate(&ctx, size - done, input_buffer + done,
		      &output_length, &output);

      if (output_length > 0)
	{
	  int res = fwrite(output, 1, output_length, stdout);
	  if (res < output_length)
	    die("Write error\n");
	}

      if (ctx.status & INFLATE_FLAG_ERROR)
	die("Invalid inflate data: %s.\n", ctx.error);	    
    }
  while (ctx.status);
  
  if (done + 4 > size)
    {      
      memmove(input_buffer, input_buffer + done, size - done);
      done = size - done;
      size = fread(input_buffer + done, 1, 4 - done, stdin);
      if (size < 4 - done)
	die("EOF in gzip trailer\n");

      trailer = input_buffer;
    }
  else
    trailer = input_buffer + done;

  process_gzip_trailer (trailer);

  return EXIT_SUCCESS;
}
